#!/usr/bin/perl  
################################################################################
#                                                                              #
#  2DHF version 2-2006                                                         #
#  Copyright (C) 2003  Jacek Kobus                                             #
#                                                                              #
#  This software may be used and distributed according to the terms            #
#  of the GNU General Public License, see README and COPYING.                  #
#                                                                              #
################################################################################

# 	$Id: elprop.pl,v 1.12 2007/01/09 14:55:19 jkob Exp $	
 
# This script facilitates the calculation of polarizabilities and
# hyperpolarizabilities from a series of calculations performed by
# means of the 2D HF program.

# Usage:
 
#    ./elprop file_m2 file_m1 file_p0 file_p1 file_p2 

# where file_i are output files from x2dhf program corresponding to
# calculations with an external finite field -2F, -F, 0, +F and +2F,
# respectively.  F is extracted from the second of these files.

# The derivatives are evaluated by means of the 3 and 5-point central
# differences formulae in order to assess the accuracy of calculations.

# For further details see 

# J.~Kobus, D.~Moncrieff, and S.~Wilson, A comparison of the electric
# moments obtained from finite basis set and finite difference
# {Hartree-Fock} calculations for diatomic molecules, Phys. Rev. A 62
# (2000), 062503/1--9.

# J.~Kobus, D.~Moncrieff, and S.~Wilson, Comparison of the
# polarizabilities and hyperpolarizabilities obtained from finite
# basis set and finite field, finite difference {Hartree-Fock}
# calculations for diatomic molecules, J. Phys. B:
# At. Mol. Opt. Phys. 34 (2001) 5127-5143.

#######################################################################

# check the number of command line arguments

foreach ($i=0; $i<=4;$i++) {
    if ( $ARGV[$i] eq '' ) {
	print ("\nError: missing argument(s) \n\n");
	print ("Usage: ./elprop file_m2 file_m1 file_p0 file_p1 file_p2 \n\n"); 
	exit;
    }
}

$debug1=0; # set to nozero value for additional printouts
$debug2=1; # set to nozero value to monitor the dependence of alpha,
           # beta and gamma (h^4) values on the accuracy of the dipole moment

$narg=5;

$F=0;
$homo="yes";
$homo="no";
$ver=0;



# extract the strength of a field and some other parameters from the
# second of the files

$x2dhf_out=$ARGV[1]; 

if ( $ARGV[5] eq "") {
    open(X2DHF_OUT,"<$x2dhf_out") || die " cannot open: $!";
    while (<X2DHF_OUT>){
	if ( $_ =~ /finite electric field:/) {
	    @t=split(" ",$_);
	    $F=abs($t[3]);
	    last;
	} 
    }
    close(X2DHF_OUT);
} 
else {

# in case the F value is not included in an output file it can be
# given as the last argument

 	$F=$ARGV[5];
# 	print "zero","\n";
    }

# extract some other parameters from the file

open(X2DHF_OUT,"<$x2dhf_out") || die " cannot open: $!";
while (<X2DHF_OUT>){
    if ( $_ =~ /nuclei/i) {
# extract Z1 Z2 
	($t[1],$Z1,$Z2)=split(" ",$_);
    }

    if ( $_ =~ /molecular system:/) {
# extract the description of a case and the internuclear distance
	$_=<X2DHF_OUT>;
	$_=<X2DHF_OUT>;
	$header=$_;
	@t=split("=",$_);
	$t[2]=$t[1];
	@t=split(" ",$t[2]);
	$R=$t[0];
    }

    if ( $_ =~ /grid:/) {
# extract the grid size
	$_=<X2DHF_OUT>;
	$_ =~ /.*\s+(\d{2,})\s+\(/;
	$nu=$1;
	$_=<X2DHF_OUT>;
	$_ =~ /\s+(\d{2,})\s+\(/;
	$mu=$1;
	$_=<X2DHF_OUT>;
	$_ =~ /\s+(\d+\.\d\d)/;
	$R_inf=$1;
    }
    if ( $_ =~ /centre\s+atomic\s+weight\s+z/ ) {
# extract atomic masses and the centre of mass
	$_=<X2DHF_OUT>;
	@t = split (" ",$_);
	$ma=$t[1];
	$_=<X2DHF_OUT>;
	@t = split (" ",$_);
	$mb=$t[1];
	$_=<X2DHF_OUT>;
	@t = split (" ",$_);
	$z=$t[1];
#	    if ($debug1) {print $ma,"  ",$mb, "  ",$z,"\n";}
    }

    if ( /electronic \(au\/Debye\-Ang\^k\)/ ) {	
	$ver=1; # output in the new format
    }
}
close(X2DHF_OUT);

# stop if the field strength has not been determined
if ($F eq "" ) {
    print ("Error: finite field strength is zero \n");
    exit;
}

if ($ver == 0) {
    extract_data_0();
} else {
    extract_data_1();
}

sub extract_data_0 {

# examine the five output files (old format; versions < 1-2005) in
# turn

    for ($iarg=0; $iarg<$narg; $iarg++) {
	
	$x2dhf_out=$ARGV[$iarg]; 
	open(X2DHF_OUT,$x2dhf_out) || die " cannot open: $!";
	
	if ($debug1) {
	    printf "\n %6s %18s \n","File: ",$x2dhf_out;
	}
	while (<X2DHF_OUT>){
	    
	    if ( $_ =~ /centre\s+atomic\s+weight\s+z/ ) {
# extract atomic masses and the centre of mass
 		for ($i=1; $i<=7; $i++) {  # skip 7 lines 
 		    $_=<X2DHF_OUT>;
 		}


# extract dipole, quadrupole etc moments 
# electronic (au)           Q_1t    Q_2t  
# electronic (Debye-Ang**k) Q_1tda  Q_2tda 
# total (au)                Q_1     Q_2
# total (Debye-Ang**k)      Q_1da   Q_2da 
		
		if ($debug1) {
		    printf "%50s \n","multipole moments relative to centre of mass:";
		}
		
		@t = split (" ",$_);
		$Q_1t[$iarg]=$t[2]; $Q_2t[$iarg]=$t[3]; $Q_3t[$iarg]=$t[4]; $Q_4t[$iarg]=$t[5];
		
		if ($debug1) {
		    printf "%33s","electronic (au)         : ";
		    for ($i=2; $i<=5; $i++) {
			printf "%15.6e", $t[$i];
		    }
		    printf "\n";
		}
		
		$_=<X2DHF_OUT>;
		@t = split (" ",$_);
		$Q_1tda[$iarg]=$t[2]; $Q_2tda[$iarg]=$t[3]; $Q_3tda[$iarg]=$t[4]; $Q_4tda[$iarg]=$t[5];
		
		if ($debug1) {
		    printf "%33s","electronic (Debye-Ang^k): ";
		    for ($i=2; $i<=5; $i++) {
			printf "%15.6e", $t[$i];
		    }
		    printf "\n";
		}
		
		$_=<X2DHF_OUT>;
		@t = split (" ",$_);
		$Q_1[$iarg]=$t[2]; $Q_2[$iarg]=$t[3]; $Q_3[$iarg]=$t[4]; $Q_4[$iarg]=$t[5];
		
		if ($debug1) {
		    printf "%33s","total (au)              : ";
		    for ($i=2; $i<=5; $i++) {
			printf "%15.6e", $t[$i];
		    }
		    printf "\n";
		}
		
		$_=<X2DHF_OUT>;
		@t = split (" ",$_);
		$Q_1da[$iarg]=$t[2]; $Q_2da[$iarg]=$t[3]; $Q_3da[$iarg]=$t[4]; $Q_4da[$iarg]=$t[5];
		
		if ($debug1) {
		    printf "%33s","total (Debye-Ang^k)     : ";
		    for ($i=2; $i<=5; $i++) {
			printf "%15.6e", $t[$i];
		    }
		    printf "\n";
		}
	    }
	    
	    if ( $_ =~ / ... writing functions to disk .../) {
# extract total energy
		$_=<X2DHF_OUT>;
		$_=<X2DHF_OUT>;
		if ( $_ =~ /total energy/) {
		    ($t[0],$t[1],$t[2]) = split (" ",$_);
		    $t[2] =~ s/D/E/;
		    $E[$iarg]=$t[2];
		    if ($debug1) {
			printf "%18s %26.16e \n","total energy:",$E[$iarg];
		    }
		}
	    }
	}
	close(X2DHF_OUT);
    }
}


sub extract_data_1 {

# examine the five output files (new format, version >= 1-2005) in turn

    for ($iarg=0; $iarg<$narg; $iarg++) {
	
	$x2dhf_out=$ARGV[$iarg]; 
	open(X2DHF_OUT,$x2dhf_out) || die " cannot open: $!";
	
	if ($debug1) {
	    printf "\n %6s %18s \n","File: ",$x2dhf_out;
	}
	
	while (<X2DHF_OUT>){

	    if ( $_ =~ /centre\s+atomic\s+weight\s+z/ ) {
# extract atomic masses and the centre of mass

		while (<X2DHF_OUT>) {  # skip some lines
		    last if (/dipole/);
		}

# extract dipole, quadrupole etc moments 
# electronic (au)           Q_1t    Q_2t  
# electronic (Debye-Ang**k) Q_1tda  Q_2tda 
# total (au)                Q_1     Q_2
# total (Debye-Ang**k)      Q_1da   Q_2da 
		
		if ($debug1) {
		    printf "%50s \n","multipole moments relative to centre of mass:";
		}

#		$_=<X2DHF_OUT>;
		@t = split (" ",$_);
		$Q_1t[$iarg]=$t[3]; $Q_1[$iarg]=$t[4];

		$_=<X2DHF_OUT>; 		
		@t = split (" ",$_);
		$Q_1tda[$iarg]=$t[0]; $Q_1da[$iarg]=$t[1];

		$_=<X2DHF_OUT>; 		
		$_=<X2DHF_OUT>; 		
		@t = split (" ",$_);
		$Q_2t[$iarg]=$t[3]; $Q_2[$iarg]=$t[4];

		$_=<X2DHF_OUT>; 		
		@t = split (" ",$_);
		$Q_2tda[$iarg]=$t[0]; $Q_2da[$iarg]=$t[1];

		$_=<X2DHF_OUT>; 		
		$_=<X2DHF_OUT>; 		
		@t = split (" ",$_);
		$Q_3t[$iarg]=$t[3]; $Q_3[$iarg]=$t[4];

		$_=<X2DHF_OUT>; 		
		@t = split (" ",$_);
		$Q_3tda[$iarg]=$t[0]; $Q_3da[$iarg]=$t[1];


		$_=<X2DHF_OUT>; 		
		$_=<X2DHF_OUT>; 		
		@t = split (" ",$_);
		$Q_4t[$iarg]=$t[3]; $Q_4[$iarg]=$t[4];

		$_=<X2DHF_OUT>; 		
		@t = split (" ",$_);
		$Q_4tda[$iarg]=$t[0]; $Q_4da[$iarg]=$t[1];

		if ($debug1) {
		    printf "%42s",  "dipole electronic total (au/Debye): ";
		    printf "%15.6e%15.6e\n", $Q_1t[$iarg],$Q_1[$iarg];
		    printf "%42s%15.6e%15.6e\n", ' ',$Q_1tda[$iarg],$Q_1da[$iarg];


		    printf "\n%42s","quadrupole electronic total (au/Debye): ";
		    printf "%15.6e%15.6e\n", $Q_2t[$iarg],$Q_2[$iarg];
		    printf "%42s%15.6e%15.6e\n", ' ', $Q_2tda[$iarg],$Q_2da[$iarg];

		    printf "\n%42s","octopole electronic total (au/Debye): ";
		    printf "%15.6e%15.6e\n", $Q_2t[$iarg],$Q_2[$iarg];
		    printf "%42s%15.6e%15.6e\n", ' ', $Q_2tda[$iarg],$Q_2da[$iarg];

		    printf "\n%42s","hexadecapole electronic total (au/Debye): ";
		    printf "%15.6e%15.6e\n", $Q_2t[$iarg],$Q_2[$iarg];
		    printf "%42s%15.6e%15.6e\n", ' ', $Q_2tda[$iarg],$Q_2da[$iarg];
		}
	    }
	    
	    if ( $_ =~ / ... writing functions to disk .../) {
# extract total energy
		$_=<X2DHF_OUT>;
		$_=<X2DHF_OUT>;
		if ( $_ =~ /total energy/) {
		    ($t[0],$t[1],$t[2]) = split (" ",$_);
		    $t[2] =~ s/D/E/;
		    $E[$iarg]=$t[2];
		    if ($debug1) {
			printf "%18s %26.16e \n","total energy:",$E[$iarg];
		    }
		}
	    }
	}
	close(X2DHF_OUT);
    }
}



# $DeltaE is the estimated relative error of the total energy.
# $DeltaQ is the (relative) error of the worst converged orbital.

# For atoms and homonuclear molecules these values can be calculated
# from differences of the total energy and dipole moment values for +/-F.

my $DeltaEdefault=1e-12;  
my $DeltaQdefault=1e-12;  


if ( ($Z1 == $Z2) or ($Z1*$Z2 == 0) ) {
    $DeltaE1= abs(($E[0]-$E[4])/$E[0]);
    $DeltaQ1= abs((abs($Q_1[0])-abs($Q_1[4]))/$Q_1[0]);

    $DeltaE2= abs(($E[1]-$E[3])/$E[1]);
    $DeltaQ2= abs((abs($Q_1[1])-abs($Q_1[3]))/$Q_1[1]);

    if ( $DeltaE1 <= $DeltaE2 ) {
	$DeltaE=$DeltaE2; 
    } else {
	$DeltaE=$DeltaE1;
    } 
 
    if ( $DeltaQ1 <= $DeltaQ2 ) {
	$DeltaQ=$DeltaQ2; 
    } else {
	$DeltaQ=$DeltaQ1;
    } 
    
#    print "+++ ",$DeltaE,"+++",$DeltaQ,"\n";
} else {
    $DeltaE=`cat DeltaE.data` || $DeltaEdefault;
    $DeltaQ=`cat DeltaQ.data` || $DeltaQdefault;
}



print "          ","\n";
print '=' x 73, "\n";
print " molecular system:          ","\n";
print ($header," \n");
print " grid:    [",$nu,"x",$mu,";",$R_inf,"]\n\n";

print " Z1        ",$Z1,"\n";
print " Z2        ",$Z2,"\n";
print " R         ",$R,"\n\n";
print " F         ",$F,"\n";

print "\n";

printf "%8s  %27.15e \n",' E(-2F) ',$E[0];
printf "%8s  %27.15e \n",' E(-1F) ',$E[1];
printf "%8s  %27.15e \n",' E( 0F) ',$E[2];
printf "%8s  %27.15e \n",' E(+1F) ',$E[3];
printf "%8s  %27.15e \n",' E(+2F) ',$E[4];

print "\n";

printf "%10s  %25.15e \n",' Q_1(-2F) ',$Q_1[0];
printf "%10s  %25.15e \n",' Q_1(-1F) ',$Q_1[1];
printf "%10s  %25.15e \n",' Q_1( 0F) ',$Q_1[2];
printf "%10s  %25.15e \n",' Q_1(+1F) ',$Q_1[3];
printf "%10s  %25.15e \n",' Q_1(+2F) ',$Q_1[4];

$changeQ1=$Q_1t[1]*(1+$DeltaQ);
$changeQ2=$Q_2t[1]*(1+$DeltaQ);
$changeE=$E[1]*(1+$DeltaE);

printf "\n";
printf "%7s %16.1e  \n",  " DeltaE",$DeltaE;
printf "%7s %11.1e  \n",  " DeltaE E(0)",abs($DeltaE*$E[2]);
printf "%7s %16.1e  \n",  " DeltaQ",$DeltaQ;
printf "\n";

# z-component of the dipole moment calculated from the 3-point formula
$mu3=-($E[3]-$E[1])/(2.0*$F)+(-($R/2.0+$z)*$Z1+($R/2.0-$z)*$Z2);

# z-component of the dipole moment calculated from the 5-point formula
$mu5=($E[4]-8*$E[3]+8*$E[1]-$E[0])/(12.0*$F)+(-($R/2.0+$z)*$Z1+($R/2.0-$z)*$Z2);

printf "%5s  %27.16f %7s %20.16f %6s \n","mu_z",$mu3,"(E h^2)",$Q_1[2],"(Q_1)    ";
printf "%5s  %27.16f %7s %20.16f %6s \n","mu_z",$mu5,"(E h^4)",$Q_1[2],"(Q_1)    ";


if ($debug2 != 0 ) {
# z-component of the dipole moment calculated from the 5-point formula

$mu5=($E[4]-8*$E[3]+8*$changeE-$E[0])/(12.0*$F)+(-($R/2.0+$z)*$Z1+($R/2.0-$z)*$Z2);
printf "%9s  %23.16f %7s %20.16f %6s \n","   +Delta",$mu5,"(E h^4)",$Q_1[2]*(1+$DeltaQ),"(Q_1)    ";
}

if ( $homo eq "yes" ) {
    $Q_1t[0] = -$Q_1t[4];
    $Q_1t[1] = -$Q_1t[3];
    $E[0]    =  $E[4];
    $E[1]    =  $E[3];
}

# alpha_zz calculated from the energy and dipole moment values by
# means of the 3-point formula (-1 0 1)

$alpha3e=-($E[3]+$E[1]-2.0*$E[2])/($F*$F);
$alpha3q= ($Q_1t[3]-$Q_1t[1])/(2.0*$F);

# alpha_zz calculated from the energy and dipole moment values by
# means of the 5-point formula (1 -8  8 -1)

$alpha5e= ($E[4]-16*$E[3]+30*$E[2]-16*$E[1]+$E[0])/(12*$F*$F);
$alpha5q= (-$Q_1t[4]+8*$Q_1t[3]-8*$Q_1t[1]+$Q_1t[0])/(12.0*$F);

print "          ","\n";
printf "%9s  %23.16f %7s %20.16f %6s \n","alpha_zz",$alpha3e,"(E h^2)",$alpha3q,"(Q_1 h^2)";
printf "%9s  %23.16f %7s %20.16f %6s \n","alpha_zz",$alpha5e,"(E h^4)",$alpha5q,"(Q_1 h^4)";

if ($debug2 != 0 ) {
#    $changeQ1=$Q_1t[1]+$DeltaQ;                
    $alpha5e= ($E[4]-16*$E[3]+30*$E[2]-16*$changeE+$E[0])/(12*$F*$F);
    $alpha5q= (-$Q_1t[4]+8*$Q_1t[3]-8*$changeQ1+$Q_1t[0])/(12.0*$F);
#    printf "%13s %49.16f %6s \n", "alpha(Delta)",$alpha5q,"(Q_1 h^4)";
    printf "%9s  %23.16f  %27.16f     \n","   +Delta",$alpha5e,$alpha5q;            
}

# zzz-component of the first hyperpolarizability calculated from the
# dipole moment values by means of the 3 and 5-point formulae (1 -2  1
# and -1 16 -30 16 -1)

$beta5e=(-$E[4]+2*$E[3]-2*$E[1]+$E[0])/(2*$F*$F*$F);
$beta3q=  ($Q_1t[3]+$Q_1t[1]-2.0*$Q_1t[2])/($F*$F);
$beta5q=  (-$Q_1t[4]+16*$Q_1t[3]-30*$Q_1t[2]+16*$Q_1t[1]-$Q_1t[0])/(12*$F*$F);

print ("\n");
printf "%9s  %52.16f %6s \n","beta_zzz",$beta3q,"(Q_1 h^2)";
printf "%9s  %23.16f %7s %20.16f %6s \n","beta_zzz",$beta5e,"(E h^4)",$beta5q,"(Q_1 h^4)";
#printf "%9s  %52.16f %6s \n","beta_zzz",$beta5q,"(Q_1 h^4)";

if ($debug2 != 0 ) {
#    $changeQ1=$Q_1t[1]+$DeltaQ;                
    $beta5e=(-$E[4]+2*$E[3]-2*$changeE+$E[0])/(2*$F*$F*$F);
    $beta5q=-($Q_1t[4]-16*$Q_1t[3]+30*$Q_1t[2]-16*$changeQ1+$Q_1t[0])/(12*$F*$F);
#    printf "%12s %50.16f %6s \n", "beta(Delta)",$beta5q,"(Q_1 h^4)";
    printf "%9s  %23.16f %7s %20.16f     \n","  +Delta",$beta5e,"(E h^4)",$beta5q;            
}

$debug3=0;
if ($debug3 != 0) {
    print ("\n");
    print ($Q_1[2],"  ",$Q_1t[2], "\n");
    print ($Q_1t[4], "   ", -$Q_1t[4]+$Q_1t[0],"\n");
    
    print ($Q_1t[3], "   ",$Q_1t[3]-$Q_1t[1]," \n");
    
    print ((-$Q_1t[4]+$Q_1t[0]), "  ",2*($Q_1t[3]-$Q_1t[1])," \n");
    
    print (((-$Q_1t[4]+$Q_1t[0])+2*($Q_1t[3]-$Q_1t[1]))/2," \n");
}

# zzzz-component of the second hyperpolarizability calculated from the
# dipole moment values by means of the 5-point formulae

$gamma5q=-(-$Q_1t[4]+2*$Q_1t[3]-2*$Q_1t[1]+$Q_1t[0])/(2*$F*$F*$F);

printf "\n";
printf "%10s  %50.16f %6s \n"," gamma_zzzz",$gamma5q,"(Q_1 h^4)";

if ($debug2 != 0 ) {
#    $changeQ1=$Q_1t[1]+$DeltaQ;                
    $gamma5q=-(-$Q_1t[4]+2*$Q_1t[3]-2*$changeQ1+$Q_1t[0])/(2*$F*$F*$F);
    printf "%11s %51.16f     \n", "     +Delta",$gamma5q;            
}


# Az,zz and Bzz,zz can be calculated as field derivatives of quadrupole moment 
$Azzz3q= ($Q_2t[3]-$Q_2t[1])/(2.0*$F);
$Azzz5q= (-$Q_2t[4]+8*$Q_2t[3]-8*$Q_2t[1]+$Q_2t[0])/(12.0*$F);
print("\n");
printf "%9s  %23.16f %7s %20.16f %6s \n","Az,zz   ",$Azzz3q ,"(  h^2)",$Azzz5q,"(Q_2 h^4)";

if ($debug2 != 0 ) {
    $Azzz5q= (-$Q_2t[4]+8*$Q_2t[3]-8*$changeQ2+$Q_2t[0])/(12.0*$F);
    printf "%11s %51.16f     \n", "     +Delta",$Azzz5q;            
}


$Bzzzz3q=  ($Q_2t[3]+$Q_2t[1]-2.0*$Q_2t[2])/($F*$F);
$Bzzzz5q=  (-$Q_2t[4]+16*$Q_2t[3]-30*$Q_2t[2]+16*$Q_2t[1]-$Q_2t[0])/(12*$F*$F);

print("\n");
printf "%9s  %23.16f %7s %20.16f %6s \n","Bzz,zz  ",$Bzzzz3q,"(  h^2)",$Bzzzz5q,"(Q_2 h^4)";

if ($debug2 != 0 ) {
    $Bzzzz5q=  (-$Q_2t[4]+16*$Q_2t[3]-30*$Q_2t[2]+16*$changeQ2-$Q_2t[0])/(12*$F*$F);
    printf "%11s %51.16f     \n", "     +Delta",$Bzzzz5q;
}


print '=' x 73, "\n";

$debug=0;

if ( $debug != 0 ) {

    print " Total multipole moments:\n";
    for ($i=0; $i<=4; $i++) {
	print '-' x 70,"\n";
#	print " File ",$i+1,"\n";
	print " Q_1  ",  $Q_1[$i], " au     ", $Q_1da[$i],  " Debye-Ang^0 \n";
	print " Q_2  ",  $Q_2[$i], " au     ", $Q_2da[$i],  " Debye-Ang^1 \n";
	print " Q_3  ",  $Q_3[$i], " au     ", $Q_3da[$i],  " Debye-Ang^2 \n";
	print " Q_4  ",  $Q_4[$i], " au     ", $Q_4da[$i],  " Debye-Ang^3 \n";
    }

    print '-' x 70,"\n";
    print "\n Electronic multipole moments:\n";
    for ($i=0; $i<=4; $i++) {
	print '-' x 70,"\n";
	print " File ",$i+1,"\n";
	print " Q_1t  ",  $Q_1[$i], " au     ", $Q_1tda[$i],  " Debye-Ang^0 \n";
	print " Q_2t  ",  $Q_2[$i], " au     ", $Q_2tda[$i],  " Debye-Ang^1 \n";
	print " Q_3t  ",  $Q_3[$i], " au     ", $Q_3tda[$i],  " Debye-Ang^2 \n";
	print " Q_4t  ",  $Q_4[$i], " au     ", $Q_4tda[$i],  " Debye-Ang^3 \n";
    }
	print '-' x 70,"\n";
}
